import math
import random
import pyglet
import const


class GameWindow(pyglet.window.Window):
    
    def __init__(self):
        super(GameWindow, self).__init__(const.WIDTH, const.HEIGHT, const.TITLE)
        self.start_intro()
    
    def start_game(self):
        self.current_window = Game(self)
        try:
            player = pyglet.media.Player()
            player.queue(pyglet.resource.media('8bp119-01-505-liberation.mp3', streaming=True))
            player.eos_action = player.EOS_LOOP
            player.play()
        except:
            print 'Unable to load music.'
    
    def start_intro(self):
        self.current_window = Intro(self)


class Intro(object):
    
    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        self.intro = pyglet.sprite.Sprite(pyglet.resource.image('intro.png'),
                                          0, 0, batch=self.batch)
        self.window.push_handlers(self.on_draw, self.on_key_press)
    
    def on_draw(self):
        self.window.clear()
        self.batch.draw()
    
    def on_key_press(self, key, modifiers):
        self.window.pop_handlers()
        self.window.start_game()


class Game(object):

    def __init__(self, window):
        self.window = window
        self.bg_batch = pyglet.graphics.Batch()
        self.batch = pyglet.graphics.Batch()
        self.gui_batch = pyglet.graphics.Batch()
        
        self.paused = False
        self.paused_label = None
        
        self.bg_sprite = pyglet.sprite.Sprite(pyglet.resource.image('bg.png'), batch=self.bg_batch, x=0, y=0)
        
        pyglet.resource.add_font('eurof35.ttf')
        pyglet.resource.add_font('eurof55.ttf')
        
        self.speed_handler = SpeedHandler()
        self.character = Character(self.batch, self.gui_batch, (100, 400))
        self.die_roller = DieRoller(self.gui_batch, (20, 20))
        self.obstacle_generator = ObstacleGenerator(self.batch, self.speed_handler)
        self.event_generator = ProbabilityEventGenerator(self.obstacle_generator, self.character)
        self.probability_display = ProbabilityDisplay(self.gui_batch, (110, 20), self.die_roller, self.event_generator)
        self.collectible_generator = CollectibleGenerator(self.batch, self.probability_display, self.speed_handler)
        self.progress_bar = ProgressBar(self.gui_batch, (20, 20))
        self.collision_handler = CollisionHandler(self.character, self.collectible_generator.collectibles,
            self.obstacle_generator.obstacles)
        self.cityline = Cityline(self.bg_batch, self.speed_handler)
        
        self.window.push_handlers(self.on_draw, self.on_key_press, self.on_key_release)
        pyglet.clock.schedule_interval(self.update, const.INV_FPS)
    
    def finish(self):
        self.window.pop_handlers()
        pyglet.clock.unschedule(self.update)
    
    def on_draw(self):
        self.window.clear()
        self.bg_batch.draw()
        self.batch.draw()
        self.gui_batch.draw()
    
    def pause(self):
        self.paused = True
        self.paused_label = pyglet.text.Label('Paused', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
    
    def unpause(self):
        self.paused = False
        self.paused_label.delete()
        self.paused_label = None
    
    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.P:
            if not self.paused:
                self.pause()
            else: self.unpause()
        elif symbol in [pyglet.window.key.LEFT, pyglet.window.key.A]:
            self.character.start_move_left()
        elif symbol in [pyglet.window.key.RIGHT, pyglet.window.key.D]:
            self.character.start_move_right()
        elif symbol in [pyglet.window.key.UP, pyglet.window.key.W]:
            self.character.start_move_up()
        elif symbol in [pyglet.window.key.DOWN, pyglet.window.key.S]:
            self.character.start_move_down()
    
    def on_key_release(self, symbol, modifiers):
        if symbol in [pyglet.window.key.LEFT, pyglet.window.key.A]:
            self.character.stop_move_left()
        elif symbol in [pyglet.window.key.RIGHT, pyglet.window.key.D]:
            self.character.stop_move_right()
        elif symbol in [pyglet.window.key.UP, pyglet.window.key.W]:
            self.character.stop_move_up()
        elif symbol in [pyglet.window.key.DOWN, pyglet.window.key.S]:
            self.character.stop_move_down()
        elif symbol == pyglet.window.key.R:
            self.finish()
            self.window.start_game()
    
    def update(self, dt):
        if self.paused: return
        self.collision_handler.update(dt)
        self.character.update(dt)
        self.collectible_generator.update(dt)
        self.probability_display.update(dt)
        self.die_roller.update(dt)
        self.obstacle_generator.update(dt)
        self.cityline.update(dt)
        self.progress_bar.update(dt)
        if self.character.dead: self.lose()
        if self.progress_bar.win: self.win()
    
    def win(self):
        self.end_label = pyglet.text.Label('The Improbability Device is fixed! You win!',
            batch=self.gui_batch, x=const.WIDTH/2, y=const.HEIGHT/2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255),
            font_name='eurofurence light', bold=True)
        self.paused = True

    def lose(self):
        self.end_label = pyglet.text.Label('Dead! :( Press R to play again.',
            batch=self.gui_batch, x=const.WIDTH/2, y=const.HEIGHT/2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255),
            font_name='eurofurence light', bold=True)
        self.paused = True


class MovingSprite(pyglet.sprite.Sprite):
    
    def __init__(self, *args, **kwargs):
        super(MovingSprite, self).__init__(*args, **kwargs)
        self.vel = [0, 0]
        self.acc = [0, 0]
    
    def update_motion(self, dt):
        self.vel[0] += self.acc[0] * dt
        self.vel[1] += self.acc[1] * dt
        self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt


class Cityline(MovingSprite):
    
    def __init__(self, batch, speed_handler):
        image = pyglet.resource.image('cityline.png')
        image.anchor_y = image.height
        super(Cityline, self).__init__(image, batch=batch)
        self.x = 0
        self.y = const.HEIGHT - 10
        self.speed_handler = speed_handler
    
    def update(self, dt):
        self.vel[0] = -self.speed_handler.speed / 4.0
        self.update_motion(dt)
        if self.x <= -const.WIDTH: self.x += const.WIDTH


class Character(MovingSprite):
    
    def __init__(self, batch, gui_batch, pos):
        image = pyglet.resource.image('character.png')
        super(Character, self).__init__(image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]
        self.speed = const.CHAR_SPEED
        self.move_left = False
        self.move_right = False
        self.move_up = False
        self.move_down = False
        self.lives = Lives(gui_batch, const.LIVES)
        self.dead = False
    
    def hit(self):
        if self.lives.decrease():
            self.dead = True
    
    def start_move_left(self):
        self.move_left = True
    
    def start_move_right(self):
        self.move_right = True

    def start_move_down(self):
        self.move_down = True

    def start_move_up(self):
        self.move_up = True

    def stop_move_left(self):
        self.move_left = False

    def stop_move_right(self):
        self.move_right = False

    def stop_move_down(self):
        self.move_down = False

    def stop_move_up(self):
        self.move_up = False
        
    def limit_boundaries(self):
        if self.x < 0: self.x = 0
        elif self.x > const.WIDTH - self.image.width: self.x = const.WIDTH - self.image.width
        if self.y < 0: self.y = 0
        elif self.y > const.HEIGHT - self.image.height: self.y = const.HEIGHT - self.image.height

    def update(self, dt):
        if self.move_left: self.vel[0] = -self.speed
        elif self.move_right: self.vel[0] = self.speed
        else: self.vel[0] = 0
        if self.move_down: self.vel[1] = -self.speed
        elif self.move_up: self.vel[1] = self.speed
        else: self.vel[1] = 0
        self.update_motion(dt)
        self.limit_boundaries()


class Lives(object):

    def __init__(self, batch, lives):
        self.batch = batch
        self.lives = lives
        self.lives_sprites = []
        self.x = 10
        self.y = 90
        self.image = pyglet.resource.image('life.png')
        self.offset = 30
        for i in xrange(lives):
            self.lives_sprites.append(pyglet.sprite.Sprite(self.image, batch=self.batch, x=self.x,
                y=self.y+i*self.offset))
    
    def decrease(self):
        self.lives -= 1
        self.lives_sprites.pop()
        if self.lives == 0:
            return True
        return False
    
    def increase(self):
        self.lives_sprites.append(pyglet.sprite.Sprite(self.image, batch=self.batch, x=self.x,
            y=self.y+self.lives*self.offset))
        self.lives += 1


class Collectible(MovingSprite):

    def __init__(self, action, speed_handler, y, *args, **kwargs):
        super(Collectible, self).__init__(*args, **kwargs)
        self.done = False
        self.action = action
        self.speed_handler = speed_handler
        self.vel[0] = -self.speed_handler.speed
        self.amplitude = random.uniform(20,60)
        self.fake_frequency_factor = random.uniform(3, 8)
        self.start_y = y
        self.timer = random.uniform(0, math.pi * 2)

    def collect(self):
        self.action()
        self.done = True

    def update(self, dt):
        self.vel[0] = -self.speed_handler.speed
        self.update_motion(dt)
        self.timer += dt * self.fake_frequency_factor
        if self.timer > 2 * math.pi: self.timer -= 2 * math.pi
        self.y = self.start_y + math.cos(self.timer) * self.amplitude
        if self.x + self.image.width < 0:
            self.done = True


class CollectibleGenerator(object):

    def __init__(self, batch, prob_disp, speed_handler):
        self.batch = batch
        self.prob_disp = prob_disp
        self.collectibles = []
        self.timer = 0
        self.speed_handler = speed_handler

    def generate(self):
        y = random.uniform(0, const.HEIGHT)
        pos = [const.WIDTH + 20, y]
        if random.randint(0, 1) == 0:
            collectible = ProbableAtom(self.batch, pos, self.prob_disp, self.speed_handler)
        else:
            collectible = ImprobableAtom(self.batch, pos, self.prob_disp, self.speed_handler)
        self.collectibles.append(collectible)

    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.timer = random.uniform(0.1, 1.0)
            self.generate()
        remove = []
        for collectible in self.collectibles:
            collectible.update(dt)
            if collectible.done:
                remove.append(collectible)
        for collectible in remove:
            self.collectibles.remove(collectible)


class ProbableAtom(Collectible):

    def __init__(self, batch, pos, prob_disp, speed_handler):
        image = pyglet.resource.image('probableAtom.png')
        super(ProbableAtom, self).__init__(prob_disp.decrement, speed_handler, pos[1], image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]


class ImprobableAtom(Collectible):

    def __init__(self, batch, pos, prob_disp, speed_handler):
        image = pyglet.resource.image('improbableAtom.png')
        super(ImprobableAtom, self).__init__(prob_disp.increment, speed_handler, pos[1], image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]


class ShuffleAtom(Collectible):

    def __init__(self, batch, pos, prob_disp):
        image = pyglet.image.SolidColorImagePattern((0,0,255,255)).create_image(10, 10)
        super(ImprobableAtom, self).__init__(prob_disp.shuffle, image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]


class ProbabilityDisplay(pyglet.text.Label):
    
    def __init__(self, batch, pos, die_roller, event_gen):
        super(ProbabilityDisplay, self).__init__('',
            batch=batch, font_size=36, anchor_x='left', anchor_y='bottom', color=(255,255,255,255),
            font_name='eurofurence light')
        self.x = pos[0]
        self.y = pos[1]
        self.die_roller = die_roller
        self.base_text = 'Nine times out of '
        self.reset_current_index()
        self.event_timer = const.SECONDS_TO_EVENT
        self.event_gen = event_gen
        self.change_event()
        self.update_text()
        self.event_indicator = EventTimerIndicator(batch, (55, 50), 50, const.SECONDS_TO_EVENT)
        self.stopped = False
        self.stopped_timer = const.DELAY_BETWEEN_EVENTS
    
    def increase(self, i):
        self.current_index += i
        if self.current_index >= len(const.PROB_VALUES):
            self.current_index = len(const.PROB_VALUES) - 1
        if self.current_index < 0:
            self.current_index = 0
    
    def increment(self):
        self.increase(1)
    
    def decrement(self):
        self.increase(-1)
    
    def shuffle(self):
        self.current_index = self.random.randint(0, len(const.PROB_VALUES) - 1)
    
    def stop_event(self):
        self.stopped = True
        self.die_roller.stopped = True
    
    def change_event(self):
        self.stopped = False
        self.event = self.event_gen.generate()
        self.reset_current_index()
        self.die_roller.reset()
    
    def reset_current_index(self):
        #self.current_index = len(const.PROB_VALUES) / 2
        self.current_index = random.randint(1, len(const.PROB_VALUES) - 2)
    
    def update(self, dt):
        self.update_text()
        if self.stopped:
            self.event_indicator.update(0)
            self.stopped_timer -= dt
            if self.stopped_timer <= 0:
                self.change_event()
                self.stopped_timer = const.DELAY_BETWEEN_EVENTS
        else:
            self.update_event_timer(dt)
            self.event_indicator.update(const.SECONDS_TO_EVENT - self.event_timer)
    
    def update_text(self):
        self.die_roller.max_value = const.PROB_VALUES[self.current_index]
        self.text = self.base_text + const.PROB_TEXTS[self.current_index] + ', ' + self.event.text
    
    def update_event_timer(self, dt):
        self.event_timer -= dt
        if self.event_timer <= 0:
            if self.die_roller.value <= 9:
                self.event.action()
            else:
                self.event.inaction()
            self.stop_event()
            self.event_timer = const.SECONDS_TO_EVENT


class ProgressBar(pyglet.sprite.Sprite):
    
    def __init__(self, batch, pos):
        image = pyglet.image.SolidColorImagePattern((0,0,255,255)).create_image(const.WIDTH - 40, 2)
        super(ProgressBar, self).__init__(image, batch=batch, x=pos[0], y=pos[1])
        indicator_image = pyglet.resource.image('progressIndicator.png')
        indicator_image.anchor_y = indicator_image.height
        self.indicator = pyglet.sprite.Sprite(indicator_image, batch=batch, x=pos[0], y=pos[1]-2)
        self.speed = 6
        self.win = False
    
    def update(self, dt):
        self.indicator.x += self.speed * dt
        if self.indicator.x + self.indicator.image.width >= const.WIDTH - 20:
            self.win = True


class DieRoller(pyglet.text.Label):

    def __init__(self, batch, pos):
        super(DieRoller, self).__init__('',
            batch=batch, font_size=56, anchor_x='left', anchor_y='bottom', color=(255,255,255,255),
            font_name='eurofurence light', bold=True)
        self.x = pos[0]
        self.y = pos[1]
        self.color_no = (128, 128, 128, 255)
        self.color_yes = (0, 255, 0, 255)
        self.max_value = 18
        self.timer = 0
        self.max_timer = 0
        self.value = random.randint(1, self.max_value)
        self.stopped = False
    
    def reset(self):
        self.timer = 0
        self.max_timer = 0
        self.stopped = False

    def update(self, dt):
        if self.stopped: return
        self.timer -= dt
        if self.timer <= 0:
            self.update_text()
            self.timer = self.max_timer
            #self.max_timer += 0.001 * (self.max_timer + 1.0)
        
    def update_text(self):
        self.value = random.randint(1, self.max_value)
        if self.value <= 9:
            self.color = self.color_yes
        else:
            self.color = self.color_no
        self.text = "%02d" % (self.value)


class EventTimerIndicator(pyglet.sprite.Sprite):

    def __init__(self, batch, pos, radius, time):
        image = pyglet.resource.image('timeIndicator.png')
        super(EventTimerIndicator, self).__init__(image, batch=batch, x=pos[0], y=pos[1]+radius)
        self.origin_x = pos[0]
        self.origin_y = pos[1]
        self.max_timer = time
        self.radius = radius
    
    def update(self, time):
        angle = 0
        if time > 0:
            angle = 2 * math.pi * (time / self.max_timer)
        self.x = self.origin_x + math.sin(angle) * self.radius
        self.y = self.origin_y + math.cos(angle) * self.radius


class CollisionHandler(object):
    
    def __init__(self, character, collectibles, obstacles):
        self.character = character
        self.collectibles = collectibles
        self.obstacles = obstacles
    
    def update(self, dt):
        for collectible in self.collectibles:
            if self.hit_sprite_box(self.character, collectible):
                collectible.collect()
        for obstacle in self.obstacles:
            if self.hit_sprite_box(self.character, obstacle):
                obstacle.hit()
                self.character.hit()
        
    def hit_sprite(self, sprite1, sprite2):
        s1x = sprite1.x - sprite1.image.anchor_x
        s2x = sprite2.x - sprite2.image.anchor_x
        s1y = sprite1.y - sprite1.image.anchor_y
        s2y = sprite2.y - sprite2.image.anchor_y
        return (not (s1x > s2x + sprite2.width or
                     s1x + sprite1.width < s2x or
                     s1y + sprite1.height < s2y or
                     s1y > s2y + sprite2.height))

    # Very quick and very stupid hack for smaller bounding box for character.
    def hit_sprite_box(self, sprite1, sprite2):
         s1x = sprite1.x - sprite1.image.anchor_x + 40
         s2x = sprite2.x - sprite2.image.anchor_x
         s1y = sprite1.y - sprite1.image.anchor_y + 10
         s2y = sprite2.y - sprite2.image.anchor_y
         return (not (s1x > s2x + sprite2.width or
                      s1x + sprite1.width - 50 < s2x or
                      s1y + sprite1.height - 20 < s2y or
                      s1y > s2y + sprite2.height))


class ProbabilityEvent(object):
    
    def __init__(self):
        self.text = ''
    
    def action(self):
        pass
    
    def inaction(self):
        pass


class EventPigsFly(ProbabilityEvent):
    
    def __init__(self, obstacle_generator):
        self.text = 'pigs fly'
        self.obstacle_generator = obstacle_generator
    
    def action(self):
        self.obstacle_generator.activate_pigs()


class EventManyPigsFly(ProbabilityEvent):

    def __init__(self, obstacle_generator):
        self.text = 'many pigs fly'
        self.obstacle_generator = obstacle_generator

    def action(self):
        self.obstacle_generator.activate_many_pigs()


class EventExtraLife(ProbabilityEvent):
    
    def __init__(self, character):
        self.text = 'a life is gained'
        self.character = character
    
    def action(self):
        self.character.lives.increase()


class EventTrafficJam(ProbabilityEvent):

    def __init__(self, obstacle_generator):
        self.text = 'a traffic jam happens'
        self.obstacle_generator = obstacle_generator

    def action(self):
        self.obstacle_generator.activate_traffic_jam()


class EventLongTrafficJam(ProbabilityEvent):

    def __init__(self, obstacle_generator):
        self.text = 'a long traffic jam happens'
        self.obstacle_generator = obstacle_generator

    def action(self):
        self.obstacle_generator.activate_long_traffic_jam()


class EventNothingBad(ProbabilityEvent):
    
    def __init__(self, obstacle_generator):
        self.text = 'nothing bad happens'
        self.obstacle_generator = obstacle_generator
    
    def inaction(self):
        v = random.randint(1, 3)
        if v == 1:
            self.obstacle_generator.activate_pigs()
        elif v == 2:
            self.obstacle_generator.activate_many_pigs()
        else:
            self.obstacle_generator.activate_traffic_jam()


class ProbabilityEventGenerator(object):
    
    def __init__(self, obstacle_generator, character):
        self.events = [EventPigsFly(obstacle_generator),
            EventExtraLife(character),
            EventTrafficJam(obstacle_generator),
            EventLongTrafficJam(obstacle_generator),
            EventManyPigsFly(obstacle_generator),
            EventNothingBad(obstacle_generator)]
    
    def generate(self):
        return random.choice(self.events)


class SpeedHandler(object):
    
    def __init__(self):
        self.speed = const.SPEED


class ObstacleGenerator(object):
    
    def __init__(self, batch, speed_handler):
        self.batch = batch
        self.speed_handler = speed_handler
        self.max_timer = 2.5
        self.timer = self.max_timer
        self.obstacles = []
        self.pig_timer = 0
        self.pig_event = False
        self.many_pig_event = False
        self.traffic_event = False
        self.pigs = 0
        self.traffic_cars = 0
    
    def generate(self):
        pos = [const.WIDTH + 30, const.HEIGHT - 40]
        fall_at_x = random.uniform(-const.WIDTH / 4, const.WIDTH)
        obstacle = Obstacle(self.batch, pos, fall_at_x, self.speed_handler)
        self.obstacles.append(obstacle)
    
    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.generate()
            self.timer = random.uniform(0.5, self.max_timer)
            if self.traffic_event:
                self.timer = 0.35
                self.traffic_cars += 1
                if self.traffic_cars > 20:
                    self.traffic_event = False
        self.update_pig(dt)
        remove = []
        for obstacle in self.obstacles:
            obstacle.update(dt)
            if obstacle.done:
                remove.append(obstacle)
        for obstacle in remove:
            self.obstacles.remove(obstacle)
    
    def activate_pigs(self):
        self.pig_event = True
        self.pigs = 0
    
    def activate_many_pigs(self):
        self.activate_pigs()
        self.many_pig_event = True
        self.pigs = -20
    
    def activate_traffic_jam(self):
        self.traffic_event = True
        self.traffic_cars = 0
    
    def activate_long_traffic_jam(self):
        self.activate_traffic_jam()
        self.traffic_cars = -20

    def generate_pig_pair(self):
        x = const.WIDTH + 30
        offset = random.uniform(0, const.HEIGHT * 0.35)
        obstacle = ObstaclePig(self.batch, [x, const.HEIGHT / 2 + offset], self.speed_handler)
        self.obstacles.append(obstacle)
        obstacle = ObstaclePig(self.batch, [x, const.HEIGHT / 2 - offset], self.speed_handler)
        self.obstacles.append(obstacle)

    def update_pig(self, dt):
        if self.pig_event:
            self.pig_timer -= dt
            if self.pig_timer <= 0:
                self.generate_pig_pair()
                self.pig_timer = 0.8
                if self.many_pig_event: self.pig_timer = 0.4
                self.pigs += 2
                if self.pigs >= 20:
                    self.pig_event = False
                    self.many_pig_event = False


class Obstacle(MovingSprite):
    
    def __init__(self, batch, pos, fall_at_x, speed_handler):
        image = pyglet.resource.image('car.png')
        super(Obstacle, self).__init__(image, batch=batch, x=pos[0], y=pos[1])
        self.speed_handler = speed_handler
        self.fall_at_x = fall_at_x
        self.done = False
    
    def fall(self):
        self.vel[1] = - const.GRAVITY
    
    def hit(self):
        self.done = True
    
    def update(self, dt):
        self.vel[0] = - self.speed_handler.speed
        self.update_motion(dt)
        if self.x < self.fall_at_x:
            self.fall()
        if self.x + self.image.width < 0 or self.y + self.image.height < 0: self.done = True


class ObstaclePig(MovingSprite):
    
    def __init__(self, batch, pos, speed_handler):
        image = pyglet.resource.image('pig.png')
        super(ObstaclePig, self).__init__(image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]
        self.vel[1] = random.uniform(-40, 40)
        self.speed_handler = speed_handler
        self.done = False

    def hit(self):
        self.done = True

    def update(self, dt):
        self.vel[0] = - self.speed_handler.speed
        self.update_motion(dt)
        if self.x + self.image.width < 0 or self.y + self.image.height < 0: self.done = True